//Ethan Alexander Shulman 2024

var pixelMap, deadPixels = [],
playerFactionId = 0, playerFaction = null, activeFactions = [],
pixelSepAng = 0, pixelCountTimer = 0;

const FACTION_COUNT = 6,
FACTIONS = [new Faction(0, 1,0,0), new Faction(1, 1,1,0), new Faction(2, 0,1,0), new Faction(3, 0,1,1), new Faction(4, 0,0,1), new Faction(5, 1,0,1)],

PIXELMAP_SIZE = 2048,
PIXELMAP_MAX = PIXELMAP_SIZE-1e-6,
PIXELMAP_CELL_SIZE = 8, PIXELMAP_CELL_SCALE = 1/PIXELMAP_CELL_SIZE,
PIXELMAP_CELLS = PIXELMAP_SIZE/PIXELMAP_CELL_SIZE,
PIXELMAP_CELLS_MINUS1 = PIXELMAP_CELLS-1,

PIXEL_RADIUS = 4, PIXEL_DIAMETER = PIXEL_RADIUS*2,
PIXEL_SPEED = 2,
PIXEL_LIFE = 8,

PIXELMAP_CELL_BORDER = 0.5-PIXEL_DIAMETER*PIXELMAP_CELL_SCALE;


function InitializePixels() {
	//init pixel map spatial look up for optimizing collisions
	pixelMap = new Array(PIXELMAP_CELLS*PIXELMAP_CELLS);
	for (var i = 0; i < pixelMap.length; i++) pixelMap[i] = [];
}


//pixel faction/team/color
/**@constructor*/
function Faction(id,r,g,b) {
	this.id = id;
	this.red = r;
	this.green = g;
	this.blue = b;
	
	this.pixels = [];
	this.pixelsBatch = new SpriteBatch();
}

//spawn pixel at point
Faction.prototype.Spawn = function(x,y) {
	var npx = new Pixel();
	npx.arrayIndex = this.pixels.push(npx)-1;
	npx.faction = this.id;
	npx.life = PIXEL_LIFE;
	
	npx.x = x;
	npx.y = y;
	npx.mapIndex = Math.floor(x*PIXELMAP_CELL_SCALE)+Math.floor(y*PIXELMAP_CELL_SCALE)*PIXELMAP_CELLS;
	npx.cellIndex = pixelMap[npx.mapIndex].push(npx)-1;
	return npx;
}


//pixel unit
/**@constructor
@extends {Sprite}*/
function Pixel() {
	Sprite.call(this);
	this.scaleX = this.scaleY = PIXEL_RADIUS*1.25;

	this.faction = 0;
	this.arrayIndex = 0;
	this.mapIndex = -1;
	this.cellIndex = -1;
	
	this.life = 0;
}
Pixel.prototype = Object.create(Sprite.prototype);


//clear all pixels
function ClearPixels() {
	for (var f = 0; f < FACTION_COUNT; f++) {
		var fpx = FACTIONS[f].pixels,
			nPx = fpx.length;
		for (var i = 0; i < nPx; i++) {
			var px = fpx[i];
			pixelMap[px.mapIndex].length = 0;
		}
		fpx.length = 0;
	}
}



//simulate pixels physics
function UpdatePixels() {
	//update spatial look up cells
	for (var f = 0; f < activeFactions.length; f++) {
		var fpx = activeFactions[f].pixels,
			pxLen = fpx.length;
		for (var i = 0; i < pxLen; i++) {
			var px = fpx[i];
			
			var nMapIndex = Math.floor(px.x*PIXELMAP_CELL_SCALE)+Math.floor(px.y*PIXELMAP_CELL_SCALE)*PIXELMAP_CELLS,
				lMapIndex = px.mapIndex;
			if (nMapIndex !== lMapIndex) {
				var oldCell = pixelMap[lMapIndex],
					endInd = oldCell.length-1,
					oldCellInd = px.cellIndex;
				if (oldCellInd !== endInd) {
					var end = oldCell[endInd];
					end.cellIndex = oldCellInd;
					oldCell[oldCellInd] = end;
				}
				oldCell.length = endInd;
				
				px.mapIndex = nMapIndex;
				px.cellIndex = pixelMap[nMapIndex].push(px)-1;
			}
		}
	}

	//simulate/update
	for (var f = 0; f < activeFactions.length; f++) {
		var fpx = activeFactions[f].pixels,
			facId = activeFactions[f].id,
			pxLen = fpx.length;
		for (var i = 0; i < pxLen; i++) {
			var px = fpx[i], x = px.x, y = px.y,
				mX = 0, mY = 0;
			
			//nearby pixel collisions
			var glX = x*PIXELMAP_CELL_SCALE, glY = y*PIXELMAP_CELL_SCALE,
				gMinX = Math.floor(glX)-1, gMinY = Math.floor(glY)-1,
				gMaxX = gMinX+3, gMaxY = gMinY+3,
				separate = false;
			gMinX = Math.max(0,gMinX); gMaxX = Math.min(PIXELMAP_CELLS,gMaxX);
			gMinY = Math.max(0,gMinY); gMaxY = Math.min(PIXELMAP_CELLS,gMaxY);
			for (var gY = gMinY; gY < gMaxY; gY++) {
				for (var gX = gMinX; gX < gMaxX; gX++) {
					var gcell = pixelMap[gX+gY*PIXELMAP_CELLS],
						cellLen = gcell.length;
					for (var c = 0; c < cellLen; c++) {
						var cPx = gcell[c];
						if (cPx === px) continue;
						
						var dX = x-cPx.x, dY = y-cPx.y,
							dLen = Math.sqrt(dX*dX+dY*dY),
							overlap = PIXEL_DIAMETER-dLen;
						if (overlap > 0) {
							if (!(cPx.faction===facId)) {
								if (--px.life === 0) deadPixels.push(px);
							}
							if (dLen === 0) {
								//pixels completely overlapped, no direction to repel by so manually separate
								separate = true;
						
							} else {
								//repel pixel
								dLen = overlap*0.5/dLen;
								mX += dX*dLen;
								mY += dY*dLen;
							}
						}
					}
				}
			}
			if (separate) {
				mX += px.separationX;
				mY += px.separationY;
			}
			
			px.x = Math.max(0,Math.min(PIXELMAP_MAX, x+Math.max(-PIXEL_SPEED,Math.min(PIXEL_SPEED, mX))));
			px.y = Math.max(0,Math.min(PIXELMAP_MAX, y+Math.max(-PIXEL_SPEED,Math.min(PIXEL_SPEED, mY))));
		}
	}
	
	var nDead = deadPixels.length;
	if (nDead) {
		//destroy dead pixels
		for (var i = 0; i < nDead; i++) {
			var dpx = deadPixels[i], dfac = dpx.faction,
				facPx = FACTIONS[dfac].pixels,
				pxId = dpx.arrayIndex, pxEnd = facPx.length-1;
			if (pxId !== pxEnd) {
				var end = facPx[pxEnd];
				end.arrayIndex = pxId;
				facPx[pxId] = end;
			}
			facPx.length = pxEnd;
		
			var cell = pixelMap[dpx.mapIndex],
				endInd = cell.length-1,
				cellInd = dpx.cellIndex;
			if (cellInd !== endInd) {
				var end = cell[endInd];
				end.cellIndex = cellInd;
				cell[cellInd] = end;
			}
			cell.length = endInd;
		}
		deadPixels.length = 0;
	}
}
